/* gInfo - Copyright (C) 2004-5 bakunin - Andrea Marchesini 
 *                                        <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include <editor.h>

enum
{
  TITLE_COLUMN,
  ARG_COLUMN,
  ITALIC_COLUMN,
  NUM_COLUMNS
};

static GtkWidget *viewer_tree;
static GtkTreeStore *viewer_model;
static GtkWidget *viewer_frame;
static GtkWidget *viewer_widget = NULL;
static GdkCursor *viewer_cursor_link = NULL;
static GdkCursor *viewer_cursor_normal = NULL;
static ginfo_docs *viewer_docs = NULL;
static int viewer_docs_n = 0;

void viewer_row_activate (GtkTreeView *, GtkTreePath *, GtkTreeViewColumn *);
void viewer_set_italic (GtkTreeView *, void *);
void viewer_container (GtkTreeView *, void *);
void viewer_container_remove (void);
GtkWidget *viewer_func (parser_info *);
GdkPixbuf *viewer_create_pixbuf (char *);
gboolean viewer_visibility_notify_event (GtkWidget *, GdkEventVisibility *,
					 gpointer);
gboolean viewer_motion_notify_event (GtkWidget *, GdkEventMotion *, gpointer);
gboolean viewer_key_press (GtkWidget *, GdkEventKey *, gpointer);
gboolean viewer_event_after (GtkWidget *, GdkEvent *, gpointer);
void viewer_check_cursor (GtkTextView *, int, int);
void viewer_go_link (GtkTextView *, GtkTextIter *);
ginfo_docs *viewer_search_link (char *);
char *viewer_trim (char *);

/* Add the viewer page in notebook */
int
viewer_notebook (GtkWidget * vbox)
{
  GtkWidget *hbox;
  GtkWidget *sw;
  GtkWidget *vp;

  GtkTreeSelection *selection;
  GtkCellRenderer *cell;
  GtkTreeViewColumn *column;

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);

  sw = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (sw);
  gtk_box_pack_start (GTK_BOX (hbox), sw, FALSE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_widget_set_size_request (sw, notebook->allocation.width / 4, -1);

  viewer_model =
    gtk_tree_store_new (NUM_COLUMNS, G_TYPE_STRING, G_TYPE_POINTER,
			G_TYPE_BOOLEAN);

  viewer_tree = gtk_tree_view_new ();
  gtk_widget_show (viewer_tree);
  gtk_tree_view_set_model (GTK_TREE_VIEW (viewer_tree),
			   GTK_TREE_MODEL (viewer_model));

  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (viewer_tree));

  gtk_container_add (GTK_CONTAINER (sw), viewer_tree);

  cell = gtk_cell_renderer_text_new ();
  g_object_set (cell, "style", PANGO_STYLE_ITALIC, NULL);

  column =
    gtk_tree_view_column_new_with_attributes ("Documentation", cell, "text",
					      TITLE_COLUMN, "style_set",
					      ITALIC_COLUMN, NULL);

  gtk_tree_view_append_column (GTK_TREE_VIEW (viewer_tree),
			       GTK_TREE_VIEW_COLUMN (column));

  g_signal_connect (viewer_tree, "row_activated",
		    G_CALLBACK (viewer_row_activate), viewer_model);

  sw = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (sw);
  gtk_box_pack_start (GTK_BOX (hbox), sw, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  vp = gtk_viewport_new (NULL, NULL);
  gtk_widget_show (vp);
  gtk_container_add (GTK_CONTAINER (sw), vp);
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (vp), GTK_SHADOW_NONE);

  viewer_frame = gtk_frame_new (NULL);
  gtk_widget_show (viewer_frame);
  gtk_container_add (GTK_CONTAINER (vp), viewer_frame);

  return 0;
}

/* Create the new tree for the parsed module */
void
viewer_create_tree (ginfo_docs * docs, int n)
{
  GtkTreeIter iter;

  viewer_docs = docs;
  viewer_docs_n = n;

  while (gtk_tree_model_iter_nth_child
	 (GTK_TREE_MODEL (viewer_model), &iter, NULL, 0))
    gtk_tree_store_remove (GTK_TREE_STORE (viewer_model), &iter);

  viewer_container_remove ();

  while (n)
    {

      gtk_tree_store_append (GTK_TREE_STORE (viewer_model), &iter, NULL);

      gtk_tree_store_set (GTK_TREE_STORE (viewer_model), &iter, TITLE_COLUMN,
			  docs->text, ARG_COLUMN, docs->arg, ITALIC_COLUMN,
			  FALSE, -1);

      docs++;
      n--;
    }

  gtk_tree_view_expand_all (GTK_TREE_VIEW (viewer_tree));
}

/* Get the click in tree list and active the correct pages */
void
viewer_row_activate (GtkTreeView * tree, GtkTreePath * path,
		     GtkTreeViewColumn * column)
{
  GtkTreeIter iter;
  GtkTreeModel *model;
  void *arg;

  model = gtk_tree_view_get_model (tree);

  gtk_tree_model_get_iter (model, &iter, path);
  gtk_tree_model_get (GTK_TREE_MODEL (model), &iter, ARG_COLUMN, &arg, -1);

  viewer_container (tree, arg);
}

/* italic... */
void
viewer_set_italic (GtkTreeView * tree, void *arg)
{
  GtkTreeIter iter;
  gboolean italic;
  GtkTreeModel *model;
  void *a;

  model = gtk_tree_view_get_model (tree);
  gtk_tree_model_get_iter_first (model, &iter);

  do
    {
      gtk_tree_model_get (GTK_TREE_MODEL (model), &iter, ITALIC_COLUMN,
			  &italic, ARG_COLUMN, &a, -1);

      if (italic && a != arg)
	gtk_tree_store_set (GTK_TREE_STORE (model), &iter, ITALIC_COLUMN, 0,
			    -1);
      else if (!italic && a == arg)
	gtk_tree_store_set (GTK_TREE_STORE (model), &iter, ITALIC_COLUMN, 1,
			    -1);

    }
  while (gtk_tree_model_iter_next (model, &iter));
}

/* attach the new page */
void
viewer_container (GtkTreeView * tree, void *arg)
{

  if (viewer_widget)
    {
      gtk_container_remove (GTK_CONTAINER (viewer_frame), viewer_widget);
    }

  viewer_widget = viewer_func ((parser_info *) arg);

  viewer_set_italic (tree, arg);

  gtk_container_add (GTK_CONTAINER (viewer_frame), viewer_widget);
}

/* remove the page */
void
viewer_container_remove (void)
{

  if (viewer_widget)
    {
      gtk_container_remove (GTK_CONTAINER (viewer_frame), viewer_widget);
    }

  viewer_widget = NULL;
}

/* This function create the page for the selected doc */
GtkWidget *
viewer_func (parser_info * info)
{
  GtkWidget *tv;
  GdkPixbuf *pixbuf;
  GtkTextBuffer *buffer;
  GtkTextTag *tag;
  GtkTextIter iter, iter_start;
  GtkTextMark *start, *end;
  parser_font font;
  int nl = 0;

  memset (&font, 0x0, sizeof (parser_font));
  font.scale = PANGO_SCALE_MEDIUM;

  tv = gtk_text_view_new ();
  gtk_widget_show (tv);

  buffer = gtk_text_buffer_new (NULL);

  viewer_cursor_link = gdk_cursor_new (GDK_HAND2);
  viewer_cursor_normal = gdk_cursor_new (GDK_XTERM);

  gtk_text_view_set_buffer (GTK_TEXT_VIEW (tv), buffer);
  gtk_container_set_border_width (GTK_CONTAINER (tv), 5);
  gtk_text_view_set_editable (GTK_TEXT_VIEW (tv), FALSE);
  gtk_text_view_set_accepts_tab (GTK_TEXT_VIEW (tv), FALSE);
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (tv), FALSE);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (tv), GTK_WRAP_WORD);

  g_signal_connect (G_OBJECT (tv), "key-press-event",
		    G_CALLBACK (viewer_key_press), tv);
  g_signal_connect (G_OBJECT (tv), "event-after",
		    G_CALLBACK (viewer_event_after), tv);
  g_signal_connect (G_OBJECT (tv), "motion-notify-event",
		    G_CALLBACK (viewer_motion_notify_event), tv);
  g_signal_connect (G_OBJECT (tv), "visibility-notify-event",
		    G_CALLBACK (viewer_visibility_notify_event), tv);

  gtk_text_buffer_get_iter_at_offset (buffer, &iter, 0);
  start = gtk_text_buffer_create_mark (buffer, "start", &iter, 1);
  end = gtk_text_buffer_create_mark (buffer, "end", &iter, 1);

  while (info)
    {
      gtk_text_buffer_move_mark (buffer, start, &iter);

      switch (info->type)
	{

	case INFO_TEXT:
	  {
	    char *last_trim;

	    if (nl)
	      last_trim = viewer_trim (info->text);
	    else
	      last_trim = info->text;

	    nl = 0;

	    gtk_text_buffer_insert (buffer, &iter, last_trim,
				    strlen (last_trim));

	    break;
	  }

	case INFO_NEWLINE:
	  nl = 1;
	  gtk_text_buffer_insert (buffer, &iter, "\n", 1);
	  break;

	case INFO_SPACE:
	  gtk_text_buffer_insert (buffer, &iter, " ", 1);
	  break;

	case INFO_IMAGE:
	  if ((pixbuf = viewer_create_pixbuf (info->image)))
	    gtk_text_buffer_insert_pixbuf (buffer, &iter, pixbuf);
	  break;


	case INFO_FONT:
	  memcpy (&font, &info->font, sizeof (parser_font));
	  break;
	}

      gtk_text_buffer_move_mark (buffer, end, &iter);

      gtk_text_buffer_get_iter_at_mark (buffer, &iter_start, start);

      /* LINK */
      if (info->link)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "foreground", "blue",
					"underline", PANGO_UNDERLINE_SINGLE,
					NULL);

	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);

	  g_object_set_data (G_OBJECT (tag), "link", info->link);
	}

      /* FONT */
      if (font.style != PANGO_STYLE_NORMAL)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "style", font.style,
					NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      if (font.variant != PANGO_VARIANT_NORMAL)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "variant", font.variant,
					NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      if (font.weight != PANGO_WEIGHT_NORMAL)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "weight", font.weight,
					NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      if (font.stretch != PANGO_STRETCH_NORMAL)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "stretch", font.stretch,
					NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      if (font.scale != PANGO_SCALE_MEDIUM)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "scale", font.scale,
					NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      /* SIZE */
      if (font.size)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "size", font.size,
					NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      /* COLOR */
      if (font.foreground)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "foreground",
					font.foreground, NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      if (font.background)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "background",
					font.background, NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      /* JUSTIFY */
      if (font.justify)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "justification",
					font.justify, NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      /* UNDERLINE */
      if (font.underline == 1)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "underline",
					PANGO_UNDERLINE_SINGLE, NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      if (font.underline == 2)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "underline",
					PANGO_UNDERLINE_DOUBLE, NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      /* FACE */
      if (font.face)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "font", font.face,
					NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      tag =
	gtk_text_buffer_create_tag (buffer, NULL, "left_margin",
				    font.margin_left, "right_margin",
				    font.margin_right, NULL);
      if (tag)
	gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);

      info = info->next;
    }

  return tv;
}

GdkPixbuf *
viewer_create_pixbuf (char *image)
{
  struct stat st;
  GdkPixbuf *pixbuf;

  if (!image)
    return NULL;

  if (!lstat (image, &st))
    {
      if ((pixbuf = gdk_pixbuf_new_from_file (image, NULL)))
	return pixbuf;

      fprintf (stderr, "I cant' convert your %s file.\n", image);
      return NULL;
    }

  if (editor_output)
    {
      char *a;
      char *b;
      int i = strlen (editor_output) - 1;

      if (!(a = strdup (editor_output)))
	fatal ("Error: memory.");

      while (i > 0)
	{
	  if (a[i] == '/')
	    break;

	  i--;
	}

      if (i > 0)
	{
	  a[i] = 0;


	  i = strlen (a) + strlen (image) + 2;
	  if (!(b = (char *) malloc (sizeof (char) * i)))
	    fatal ("Error: memory.");

	  snprintf (b, i, "%s/%s", a, image);

	  if (!lstat (b, &st))
	    {
	      if ((pixbuf = gdk_pixbuf_new_from_file (b, NULL)))
		return pixbuf;

	      fprintf (stderr, "I cant' convert your %s file.\n", b);
	      return NULL;
	    }

	  free (b);

	}

      free (a);
    }

  fprintf (stderr, "Image %s not found!\n", image);
  return NULL;
}

char *
viewer_trim (char *text)
{
  int i = 0, len;
  int j = 0;

  len = strlen (text);

  while (i < len)
    {
      if (text[i++] != ' ')
	break;
      else
	j++;
    }

  return &text[j];
}

gboolean
viewer_key_press (GtkWidget * w, GdkEventKey * event, gpointer dummy)
{
  GtkTextIter iter;
  GtkTextBuffer *buffer;

  switch (event->keyval)
    {
    case GDK_Return:
    case GDK_KP_Enter:
      buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (dummy));
      gtk_text_buffer_get_iter_at_mark (buffer, &iter,
					gtk_text_buffer_get_insert (buffer));
      viewer_go_link (GTK_TEXT_VIEW (dummy), &iter);
      break;

    default:
      break;
    }

  return FALSE;

}

gboolean
viewer_visibility_notify_event (GtkWidget * w, GdkEventVisibility * event,
				gpointer dummy)
{
  gint wx, wy, bx, by;

  gdk_window_get_pointer (GTK_WIDGET (dummy)->window, &wx, &wy, NULL);

  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (dummy),
					 GTK_TEXT_WINDOW_WIDGET, wx, wy,
					 &bx, &by);

  viewer_check_cursor (GTK_TEXT_VIEW (dummy), bx, by);
  gdk_window_get_pointer (GTK_WIDGET (dummy)->window, NULL, NULL, NULL);

  return FALSE;
}

gboolean
viewer_motion_notify_event (GtkWidget * w, GdkEventMotion * event,
			    gpointer dummy)
{
  gint x, y;
  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (dummy),
					 GTK_TEXT_WINDOW_WIDGET, event->x,
					 event->y, &x, &y);

  viewer_check_cursor (GTK_TEXT_VIEW (dummy), x, y);
  gdk_window_get_pointer (GTK_WIDGET (dummy)->window, NULL, NULL, NULL);

  return FALSE;
}

gboolean
viewer_event_after (GtkWidget * w, GdkEvent * event, gpointer dummy)
{
  GtkTextIter start, end, iter;
  GtkTextBuffer *buffer;
  GdkEventButton *e;
  gint x, y;

  if (event->type != GDK_BUTTON_RELEASE)
    return FALSE;

  e = (GdkEventButton *) event;

  if (e->button != 1)
    return FALSE;

  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (dummy));

  gtk_text_buffer_get_selection_bounds (buffer, &start, &end);
  if (gtk_text_iter_get_offset (&start) != gtk_text_iter_get_offset (&end))
    return FALSE;

  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (dummy),
					 GTK_TEXT_WINDOW_WIDGET, e->x, e->y,
					 &x, &y);

  gtk_text_view_get_iter_at_location (GTK_TEXT_VIEW (dummy), &iter, x, y);
  viewer_go_link (GTK_TEXT_VIEW (dummy), &iter);

  return FALSE;
}

void
viewer_check_cursor (GtkTextView * tv, int x, int y)
{
  GSList *tags = NULL, *tagb = NULL;
  GtkTextBuffer *buffer;
  GtkTextIter iter;
  int do_link = 0;

  buffer = gtk_text_view_get_buffer (tv);
  gtk_text_view_get_iter_at_location (tv, &iter, x, y);
  tags = tagb = gtk_text_iter_get_tags (&iter);

  while (tags)
    {
      GtkTextTag *tag = tags->data;
      char *link = (char *) g_object_get_data (G_OBJECT (tag), "link");

      if (link)
	do_link = 1;

      tags = tags->next;
    }

  if (do_link)
    gdk_window_set_cursor (gtk_text_view_get_window
			   (tv, GTK_TEXT_WINDOW_TEXT), viewer_cursor_link);
  else
    gdk_window_set_cursor (gtk_text_view_get_window
			   (tv, GTK_TEXT_WINDOW_TEXT), viewer_cursor_normal);

  if (tagb)
    g_slist_free (tagb);
}

void
viewer_go_link (GtkTextView * tv, GtkTextIter * iter)
{
  GSList *tags = NULL, *tagb = NULL;

  tags = tagb = gtk_text_iter_get_tags (iter);

  while (tags)
    {
      GtkTextTag *tag = tags->data;
      char *link = g_object_get_data (G_OBJECT (tag), "link");

      if (link != 0)
	{
	  ginfo_docs *d = viewer_search_link (link);

	  if (d)
	    viewer_container (GTK_TREE_VIEW (viewer_tree), d->arg);
	  else
	    fprintf (stderr, "I can't find %s link!\n", link);

	  break;
	}
      tags = tags->next;
    }

  if (tagb)
    g_slist_free (tagb);
}

ginfo_docs *
viewer_search_link (char *link)
{
  ginfo_docs *d = viewer_docs;
  int n = 0;

  while (n < viewer_docs_n)
    {
      if (d[n].id)
	if (!strcmp (link, d[n].id))
	  return &d[n];

      n++;
    }

  return NULL;
}

/* EOF */
