/* gInfo - Copyright (C) 2004-5 bakunin - Andrea Marchesini 
 *                                        <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include "ginfo.h"

ginfo_docs *modules_ginfo_check_docs (xmlDocPtr, xmlNodePtr, int *);
ginfo_tips *modules_ginfo_check_tips (xmlDocPtr, xmlNodePtr, int *);
module_xml_info *modules_ginfo_check_info (xmlDocPtr, xmlNodePtr);
int modules_ginfo_check_info_tag (xmlDocPtr, xmlNodePtr, char *,
				  module_xml_info_font *);
void modules_ginfo_check_info_rec (xmlDocPtr, xmlNodePtr, module_xml_info **,
				   module_xml_info **,
				   module_xml_info_font *);
void modules_ginfo_null (void *, const char *, ...);

int
modules_ginfo_check (xmlDocPtr doc, xmlNodePtr cur, char *a)
{
  module *m;
  char *def = NULL;
  char *name = NULL;
  char *genre = NULL;
  char *words = NULL;
  ginfo_tips *tips = NULL;
  ginfo_docs *docs = NULL;
  int docs_n = 0;
  int tips_n = 0;

  if (!(m = (module *) malloc (sizeof (module))))
    fatal ("Error: memory.");

  m->type = GINFO_XML;

  for (cur = cur->children; cur != NULL; cur = cur->next)
    {
      if (!xmlStrcmp (cur->name, "module"))
	{
	  def = xmlGetProp (cur, "default");

	  if (!(name = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*name)
	    {
	      msg ("Module %s: no name.", a);
	      free (m);
	      return 1;
	    }

	  name = trim (name);

	  if (def)
	    def = trim (def);
	}

      else if (!xmlStrcmp (cur->name, "genre"))
	{
	  if (!(genre = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*genre)
	    {
	      msg ("Module %s: no genre.", a);
	      free (m);
	      return 1;
	    }

	  genre = trim (genre);
	}

      else if (!xmlStrcmp (cur->name, "words"))
	{
	  if (!(words = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*words)
	    {
	      msg ("Module %s: no words.", a);
	      free (m);
	      return 1;
	    }

	  words = trim (words);
	}

      else if (!xmlStrcmp (cur->name, "tooltips"))
	{
	  if (!(tips = modules_ginfo_check_tips (doc, cur, &tips_n)))
	    {
	      msg ("Module %s: tips syntax error.", a);
	      free (m);
	      return 1;
	    }
	}

      else if (!xmlStrcmp (cur->name, "docs"))
	{
	  if (!(docs = modules_ginfo_check_docs (doc, cur, &docs_n)))
	    {
	      msg ("Module %s: docs syntax error.", a);
	      free (m);
	      return 1;
	    }
	}
    }

  m->handle = doc;

  if (!name)
    {
      msg ("Module %s: no name.", a);

      free (m);
      return 1;
    }

  m->name = name;

  if (!genre)
    {
      msg ("Module %s: no genre.", a);
      free (m);
      return 1;
    }

  if (genre && !(m->genre = ginfo_check_genre (genre)))
    {
      msg ("Module %s: no compatible genre.", a);
      free (m);
      return 1;
    }

  if (!words)
    {
      msg ("Module %s: no words.", a);
      free (m);
      return 1;
    }

  if (!(m->words = ginfo_check_words (words)))
    {
      msg ("Module %s: no compatible words.", a);
      free (m);
      return 1;
    }

  m->docs = NULL;

  if (!docs)
    {
      msg ("Module %s: no docs.", a);
      free (m->words);
      free (m);
      return 1;
    }

  if (docs_add (docs, docs_n, m))
    {
      msg ("Module %s: no docs.", a);
      ginfo_easy_docs_list_free (docs);
      free (m->words);
      free (m);
      return 1;
    }

  if (tips)
    tooltips_add (tips, tips_n, m);

  if (!(m->file = strdup (a)))
    fatal ("Error: memory");

  m->next = module_head;
  module_head = m;

  if (def && *def == '1')
    modules_set_default (a, m);

  return 0;
}

ginfo_docs *
modules_ginfo_check_docs (xmlDocPtr doc, xmlNodePtr c, int *n)
{
  void *t = NULL;
  xmlNodePtr cur, cur_c;
  char *chapter;
  module_xml_info *info;
  char *id;

  for (cur = c->children; cur != NULL; cur = cur->next)
    {
      if (!xmlStrcmp (cur->name, "doc"))
	{

	  id = xmlGetProp (cur, "id");
	  if (!id || !*id)
	    id = NULL;

	  if (!t)
	    t = ginfo_easy_docs_init ();

	  for (cur_c = cur->children; cur_c != NULL; cur_c = cur_c->next)
	    {
	      if (!xmlStrcmp (cur_c->name, "chapter"))
		{
		  if (!
		      (chapter =
		       xmlNodeListGetString (doc, cur_c->xmlChildrenNode, 1))
		      || !*chapter)
		    {
		      ginfo_easy_docs_free (t);
		      return NULL;
		    }

		  chapter = trim (chapter);

		}
	      else if (!xmlStrcmp (cur_c->name, "info"))
		{
		  if (!
		      (info =
		       modules_ginfo_check_info (doc, cur_c->children)))
		    {
		      ginfo_easy_docs_free (t);
		      return NULL;
		    }

		}
	    }

	  ginfo_easy_docs_add (t, chapter, ginfo_tutor_xml, (void *) info,
			       id);
	}
    }

  if (!t)
    return NULL;

  return ginfo_easy_docs_close (t, n);
}

ginfo_tips *
modules_ginfo_check_tips (xmlDocPtr doc, xmlNodePtr c, int *n)
{
  xmlNodePtr cur;
  char *tip;
  void *t = NULL;
  char *id;

  for (cur = c->children; cur != NULL; cur = cur->next)
    {
      if (!xmlStrcmp (cur->name, "tip"))
	{
	  if (!
	      (tip =
	       xmlNodeListGetString (doc, cur->xmlChildrenNode, 1)) || !*tip)
	    return NULL;

	  tip = trim (tip);

	  id = xmlGetProp (cur, "id");
	  if (!id || !*id)
	    id = NULL;

	  if (!t)
	    t = ginfo_easy_tips_init ();

	  ginfo_easy_tips_add (t, tip, NULL, NULL, id);
	}
    }

  if (!t)
    return NULL;

  return ginfo_easy_tips_close (t, n);
}

module_xml_info *
modules_ginfo_check_info (xmlDocPtr doc, xmlNodePtr cur)
{
  module_xml_info *t = NULL;
  module_xml_info *h = NULL;
  module_xml_info_font font;

  memset (&font, 0x0, sizeof (font));
  font.scale = PANGO_SCALE_MEDIUM;

  modules_ginfo_check_info_rec (doc, cur, &t, &h, &font);

  return h;
}

void
modules_ginfo_check_info_rec (xmlDocPtr doc, xmlNodePtr cur,
			      module_xml_info ** t, module_xml_info ** h,
			      module_xml_info_font * font)
{

  module_xml_info *c;
  char *link = NULL;

  if (*t)
    link = (*t)->link;

  while (cur)
    {

      c = NULL;

      if (cur->type == XML_TEXT_NODE)
	{
	  if (!(c = (module_xml_info *) malloc (sizeof (module_xml_info))))
	    fatal ("Error: memory.");

	  c->next = NULL;
	  c->link = link;
	  c->image = NULL;
	  c->type = INFO_TEXT;

	  c->text = xmlNodeGetContent (cur);
	  if (c->text)
	    c->text = modules_xml_info_trim (c->text);

	  if (!c->text)
	    {
	      free (c);
	      c = NULL;
	    }
	}
      else if (cur->type == XML_ELEMENT_NODE)
	{
	  if (!(c = (module_xml_info *) malloc (sizeof (module_xml_info))))
	    fatal ("Error: memory.");

	  c->next = NULL;
	  c->text = NULL;
	  c->link = link;
	  c->image = NULL;

	  if (!strcmp (cur->name, "nl"))
	    c->type = INFO_NEWLINE;

	  else if (!strcmp (cur->name, "space"))
	    c->type = INFO_SPACE;

	  else if (!strcmp (cur->name, "image"))
	    {
	      c->type = INFO_IMAGE;

	      if (!
		  (c->image =
		   xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
		  || !*c->image)
		{
		  free (c);
		  c = NULL;
		}
	    }

	  else
	    {
	      c->type = INFO_FONT;

	      memcpy (&c->font, font, sizeof (module_xml_info_font));

	      if (!strcmp (cur->name, "link"))
		{
		  c->link = xmlGetProp (cur, "id");
		  if (!c->link || !*c->link)
		    {
		      msg ("Axpect tag link id=???");
		      free (c);
		      c = NULL;
		    }
		}

	      else if (modules_ginfo_check_info_tag
		       (doc, cur, (char *) cur->name, &c->font))
		{
		  free (c);
		  c = NULL;
		}
	    }
	}

      if (c)
	{
	  if (!*h)
	    *h = c;
	  else
	    (*t)->next = c;

	  *t = c;

	  if (c->type == INFO_FONT)
	    {
	      modules_ginfo_check_info_rec (doc, cur->children, t, h,
					    &c->font);
	      if (!
		  (c = (module_xml_info *) malloc (sizeof (module_xml_info))))
		fatal ("Error: memory.");

	      c->next = NULL;
	      c->text = NULL;
	      (*t)->next = c;
	      *t = c;

	      c->type = INFO_FONT;

	      memcpy (&c->font, font, sizeof (module_xml_info_font));
	    }
	}

      cur = cur->next;
    }
}

int
modules_ginfo_check_info_tag (xmlDocPtr doc, xmlNodePtr cur, char *tag,
			      module_xml_info_font * font)
{
  /* STYLE */
  if (!strcmp (tag, "oblique"))
    {
      font->style = PANGO_STYLE_OBLIQUE;
      return 0;
    }

  if (!strcmp (tag, "italic"))
    {
      font->style = PANGO_STYLE_ITALIC;
      return 0;
    }

  /* VARIANT */
  if (!strcmp (tag, "small_caps"))
    {
      font->variant = PANGO_VARIANT_SMALL_CAPS;
      return 0;
    }

  /* WEIGHT */
  if (!strcmp (tag, "ultralight"))
    {
      font->weight = PANGO_WEIGHT_ULTRALIGHT;
      return 0;
    }

  if (!strcmp (tag, "light"))
    {
      font->weight = PANGO_WEIGHT_LIGHT;
      return 0;
    }

  if (!strcmp (tag, "bold"))
    {
      font->weight = PANGO_WEIGHT_BOLD;
      return 0;
    }

  if (!strcmp (tag, "ultrabold"))
    {
      font->weight = PANGO_WEIGHT_ULTRABOLD;
      return 0;
    }

  if (!strcmp (tag, "heavy"))
    {
      font->weight = PANGO_WEIGHT_HEAVY;
      return 0;
    }

  /* STRETCH */
  if (!strcmp (tag, "ultracondensed"))
    {
      font->stretch = PANGO_STRETCH_ULTRA_CONDENSED;
      return 0;
    }

  if (!strcmp (tag, "extracondensed"))
    {
      font->stretch = PANGO_STRETCH_EXTRA_CONDENSED;
      return 0;
    }

  if (!strcmp (tag, "condensed"))
    {
      font->stretch = PANGO_STRETCH_CONDENSED;
      return 0;
    }

  if (!strcmp (tag, "semicondensed"))
    {
      font->stretch = PANGO_STRETCH_SEMI_CONDENSED;
      return 0;
    }

  if (!strcmp (tag, "semiexpanded"))
    {
      font->stretch = PANGO_STRETCH_SEMI_EXPANDED;
      return 0;
    }

  if (!strcmp (tag, "expanded"))
    {
      font->stretch = PANGO_STRETCH_EXPANDED;
      return 0;
    }

  if (!strcmp (tag, "extraexpanded"))
    {
      font->stretch = PANGO_STRETCH_EXTRA_EXPANDED;
      return 0;
    }

  if (!strcmp (tag, "ultraexpanded"))
    {
      font->stretch = PANGO_STRETCH_ULTRA_EXPANDED;
      return 0;
    }

  /* SCALE */
  if (!strcmp (tag, "xx_small"))
    {
      font->scale = PANGO_SCALE_XX_SMALL;
      return 0;
    }

  if (!strcmp (tag, "x_small"))
    {
      font->scale = PANGO_SCALE_X_SMALL;
      return 0;
    }

  if (!strcmp (tag, "small"))
    {
      font->scale = PANGO_SCALE_SMALL;
      return 0;
    }

  if (!strcmp (tag, "large"))
    {
      font->scale = PANGO_SCALE_LARGE;
      return 0;
    }

  if (!strcmp (tag, "x_large"))
    {
      font->scale = PANGO_SCALE_X_LARGE;
      return 0;
    }

  if (!strcmp (tag, "xx_large"))
    {
      font->scale = PANGO_SCALE_XX_LARGE;
      return 0;
    }

  /* SIZE */
  if (!strcmp (tag, "size"))
    {
      char *size;
      size = xmlGetProp (cur, "value");
      if (!size || !*size)
	{
	  msg ("Axpect tag %s value=???", tag);
	  return 1;
	}

      font->size = atoi (size) * PANGO_SCALE;

      return 0;
    }

  /* COLOR */
  if (!strcmp (tag, "foreground"))
    {
      font->foreground = xmlGetProp (cur, "value");
      if (!font->foreground || !*font->foreground)
	{
	  msg ("Axpect tag %s value=???", tag);
	  return 1;
	}

      return 0;
    }

  if (!strcmp (tag, "background"))
    {
      font->background = xmlGetProp (cur, "value");
      if (!font->background || !*font->background)
	{
	  msg ("Axpect tag %s value=???", tag);
	  return 1;
	}

      return 0;
    }

  /* JUSTIFY */
  if (!strcmp (tag, "justify"))
    {
      char *justify;
      justify = xmlGetProp (cur, "value");

      if (!justify || !*justify)
	{
	  msg ("Axpect tag %s value=???", tag);
	  return 1;
	}

      if (!strcmp (justify, "center"))
	font->justify = GTK_JUSTIFY_CENTER;
      else if (!strcmp (justify, "left"))
	font->justify = GTK_JUSTIFY_LEFT;
      else if (!strcmp (justify, "right"))
	font->justify = GTK_JUSTIFY_RIGHT;
      else if (!strcmp (justify, "fill"))
	font->justify = GTK_JUSTIFY_FILL;
      else
	{
	  msg ("Axpect tag %s:value center|left|center", tag);
	  return 1;
	}

      return 0;
    }

  /* UNDERLINE */
  if (!strcmp (tag, "underline"))
    {
      font->underline = 1;
      return 0;
    }

  if (!strcmp (tag, "double_underline"))
    {
      font->underline = 2;
      return 0;
    }

  /* FACE */
  if (!strcmp (tag, "face"))
    {
      font->face = xmlGetProp (cur, "value");
      if (!font->face || !*font->face)
	{
	  msg ("Axpect tag %s value=???", tag);
	  return 1;
	}

      return 0;
    }

  /* MARGIN */
  if (!strcmp (tag, "margin"))
    {
      char *v;

      v = xmlGetProp (cur, "left");
      if (v)
	{
	  font->margin_left = atoi (v);
	  if (font->margin_left < 0)
	    font->margin_left = 0;
	}

      v = xmlGetProp (cur, "right");
      if (v)
	{
	  font->margin_right = atoi (v);
	  if (font->margin_right < 0)
	    font->margin_right = 0;
	}

      return 0;
    }

  /* FONT */
  if (!strcmp (tag, "font"))
    {
      char *tmp;
      tmp = xmlGetProp (cur, "style");

      /* Style */
      if (tmp)
	{
	  if (!strcmp (tmp, "oblique"))
	    font->style = PANGO_STYLE_OBLIQUE;
	  else if (!strcmp (tmp, "italic"))
	    font->style = PANGO_STYLE_ITALIC;
	  else
	    {
	      msg ("Font:Style axpected oblique|italic");
	      return 1;
	    }
	}

      /* Variant */
      tmp = xmlGetProp (cur, "variant");
      if (tmp)
	{
	  if (!strcmp (tmp, "small_cups"))
	    font->variant = PANGO_VARIANT_SMALL_CAPS;
	  else
	    {
	      msg ("Font:Variant axpected small_caps");
	      return 1;
	    }
	}

      /* Weight */
      tmp = xmlGetProp (cur, "weight");
      if (tmp)
	{
	  if (!strcmp (tmp, "ultralight"))
	    font->weight = PANGO_WEIGHT_ULTRALIGHT;
	  else if (!strcmp (tmp, "light"))
	    font->weight = PANGO_WEIGHT_LIGHT;
	  else if (!strcmp (tmp, "bold"))
	    font->weight = PANGO_WEIGHT_BOLD;
	  else if (!strcmp (tmp, "ultrabold"))
	    font->weight = PANGO_WEIGHT_ULTRABOLD;
	  else if (!strcmp (tmp, "heavy"))
	    font->weight = PANGO_WEIGHT_HEAVY;
	  else
	    {
	      msg
		("Font:Weight axpected ultralight|light|bold|ultrabold|heavy");
	      return 1;
	    }
	}

      /* Stretch */
      tmp = xmlGetProp (cur, "stretch");
      if (tmp)
	{
	  if (!strcmp (tmp, "ultracondensed"))
	    font->stretch = PANGO_STRETCH_ULTRA_CONDENSED;
	  else if (!strcmp (tmp, "extracondensed"))
	    font->stretch = PANGO_STRETCH_EXTRA_CONDENSED;
	  else if (!strcmp (tmp, "condensed"))
	    font->stretch = PANGO_STRETCH_CONDENSED;
	  else if (!strcmp (tmp, "semicondensed"))
	    font->stretch = PANGO_STRETCH_SEMI_CONDENSED;
	  else if (!strcmp (tmp, "semiexpanded"))
	    font->stretch = PANGO_STRETCH_SEMI_EXPANDED;
	  else if (!strcmp (tmp, "expanded"))
	    font->stretch = PANGO_STRETCH_EXPANDED;
	  else if (!strcmp (tmp, "extraexpanded"))
	    font->stretch = PANGO_STRETCH_EXTRA_EXPANDED;
	  else if (!strcmp (tmp, "ultraexpanded"))
	    font->stretch = PANGO_STRETCH_ULTRA_EXPANDED;
	  else
	    {
	      msg
		("Font:Stretch axpected ultracondensed|extracondensed|condensed|semicondensed|semiexpanded|expanded|extraexpanded|ultraexpanded");
	      return 1;
	    }
	}

      /* Scale */
      tmp = xmlGetProp (cur, "scale");
      if (tmp)
	{
	  if (!strcmp (tmp, "xx_small"))
	    font->scale = PANGO_SCALE_XX_SMALL;
	  if (!strcmp (tmp, "x_small"))
	    font->scale = PANGO_SCALE_X_SMALL;
	  if (!strcmp (tmp, "small"))
	    font->scale = PANGO_SCALE_SMALL;
	  if (!strcmp (tmp, "large"))
	    font->scale = PANGO_SCALE_LARGE;
	  if (!strcmp (tmp, "x_large"))
	    font->scale = PANGO_SCALE_X_LARGE;
	  if (!strcmp (tmp, "xx_large"))
	    font->scale = PANGO_SCALE_XX_LARGE;
	  else
	    {
	      msg
		("Font:Scale axpected xx_small|x_small|small|large|x_lage|xx_large");
	      return 1;
	    }
	}

      /* Size */
      tmp = xmlGetProp (cur, "size");
      if (tmp)
	{
	  font->size = atoi (tmp) * PANGO_SCALE;
	}

      /* Color */
      font->foreground = xmlGetProp (cur, "foreground");
      font->background = xmlGetProp (cur, "background");

      /* Justify */
      tmp = xmlGetProp (cur, "justify");
      if (tmp)
	{
	  if (!strcmp (tmp, "center"))
	    font->justify = GTK_JUSTIFY_CENTER;
	  else if (!strcmp (tmp, "left"))
	    font->justify = GTK_JUSTIFY_LEFT;
	  else if (!strcmp (tmp, "right"))
	    font->justify = GTK_JUSTIFY_RIGHT;
	  else
	    {
	      msg ("Font:Justify axpected center|left|right");
	      return 1;
	    }
	}

      /* Underline */
      tmp = xmlGetProp (cur, "underline");
      if (tmp)
	{
	  if (!strcmp (tmp, "single"))
	    font->underline = 1;
	  else if (!strcmp (tmp, "double"))
	    font->underline = 2;
	  else
	    {
	      msg ("Font:Underline axpected single|double");
	      return 1;
	    }
	}

      /* Face */
      font->face = xmlGetProp (cur, "face");

      /* margin_left */
      tmp = xmlGetProp (cur, "margin_left");
      if (tmp)
	{
	  font->margin_left = atoi (tmp);
	  if (font->margin_left < 0)
	    font->margin_left = 0;
	}

      /* margin_right */
      tmp = xmlGetProp (cur, "margin_right");
      if (tmp)
	{
	  font->margin_right = atoi (tmp);
	  if (font->margin_right < 0)
	    font->margin_right = 0;
	}

      return 0;
    }

  msg ("Unknow tag %s!\n", tag);
  return 1;
}

/* EOF */
